from django.contrib.auth import get_user_model
from rest_framework import serializers
from .models import *

###
# Users.

class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ['username', 'email', 'balance']

    # Display current 'balance'
    balance = serializers.SerializerMethodField('get_balance')

    def get_balance(self, user):
        return get_balance_for(user)

###
# Payments and Payment Methods.

class PaymentSerializer(serializers.ModelSerializer):
    class Meta:
        model = Payment
        fields = ['owner', 'amount', 'source', 'timestamp']

class PaymentMethodSerializer(serializers.ModelSerializer):
    stripe_card_last4 = serializers.IntegerField()

    class Meta:
        model = PaymentMethod
        fields = ['uuid', 'source', 'description', 'primary', 'stripe_card_last4']

class ChargePaymentMethodSerializer(serializers.Serializer):
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)

class CreditCardSerializer(serializers.Serializer):
    number = serializers.IntegerField()
    exp_month = serializers.IntegerField()
    exp_year = serializers.IntegerField()
    cvc = serializers.IntegerField()

class CreatePaymentMethodSerializer(serializers.ModelSerializer):
    credit_card = CreditCardSerializer()

    class Meta:
        model = PaymentMethod
        fields = ['source', 'description', 'primary', 'credit_card']

###
# Orders & Products.

class OrderRecordSerializer(serializers.ModelSerializer):
    class Meta:
        model = OrderRecord
        fields  = ['one_time_price', 'recurring_price', 'description']


class OrderSerializer(serializers.ModelSerializer):
    records = OrderRecordSerializer(many=True, read_only=True)
    class Meta:
        model = Order
        fields = ['uuid', 'creation_date', 'starting_date', 'ending_date',
                'bill', 'recurring_period', 'records', 'recurring_price', 'one_time_price']


###
# Bills

# TODO: remove magic numbers for decimal fields
class BillRecordSerializer(serializers.Serializer):
    order = serializers.HyperlinkedRelatedField(
            view_name='order-detail',
            read_only=True)
    description = serializers.CharField()
    recurring_period = serializers.CharField()
    recurring_price = serializers.DecimalField(max_digits=10, decimal_places=2)
    recurring_count = serializers.DecimalField(max_digits=10, decimal_places=2)
    one_time_price = serializers.DecimalField(max_digits=10, decimal_places=2)
    amount = serializers.DecimalField(max_digits=10, decimal_places=2)

class BillSerializer(serializers.ModelSerializer):
    records = BillRecordSerializer(many=True, read_only=True)
    class Meta:
        model = Bill
        fields = ['reference', 'owner', 'total', 'due_date', 'creation_date',
                'starting_date', 'ending_date', 'records', 'final']
